# ------------------------------------------------------------------------------
#  mainframe.tcl
#  This file is part of Unifix BWidget Toolkit
# ------------------------------------------------------------------------------
#  Index of commands:
#     - MainFrame::create
#     - MainFrame::configure
#     - MainFrame::cget
#     - MainFrame::getframe
#     - MainFrame::addtoolbar
#     - MainFrame::gettoolbar
#     - MainFrame::addindicator
#     - MainFrame::getindicator
#     - MainFrame::showtoolbar
#     - MainFrame::showstatusbar
#     - MainFrame::_create_menubar
#     - MainFrame::_create_entries
#     - MainFrame::_parse_name
#     - MainFrame::_parse_accelerator
# ------------------------------------------------------------------------------

namespace eval MainFrame {
    Widget::declare MainFrame {
        {-width        TkResource 0      0 frame}
        {-height       TkResource 0      0 frame}
        {-textvariable String     ""     0}
        {-variable     String     ""     0}
        {-menu         String     ""     1}
        {-separator    Enum       both   1 {none top bottom both}}
    }

    Widget::addmap MainFrame "" .user {-width {} -height {}}

    proc ::MainFrame { path args } { return [eval MainFrame::create $path $args] }
    proc use {} {}

    variable _widget
}


# ------------------------------------------------------------------------------
#  Command MainFrame::create
# ------------------------------------------------------------------------------
proc MainFrame::create { path args } {
    global   tcl_platform
    variable _widget

    frame $path
    set top [winfo parent $path]
    if { [string compare [winfo toplevel $path] $top] } {
        destroy $path
        return -code error "parent must be a toplevel"
    }
    Widget::init MainFrame $path $args

    if { $tcl_platform(platform) == "unix" } {
        set relief raised
        set bd     1
    } else {
        set relief flat
        set bd     0
    }
    set topframe  [frame $path.topf -relief flat -borderwidth 0]
    set userframe [eval frame $path.user [Widget::subcget $path .user] \
                       -relief $relief -borderwidth $bd]
    set botframe  [frame $path.botf -relief $relief -borderwidth $bd]

    pack $topframe -fill x
    grid columnconfigure $topframe 0 -weight 1

    if { $tcl_platform(platform) != "unix" } {
        set sepopt [Widget::cget $path -separator]
        if { $sepopt == "both" || $sepopt == "top" } {
            set sep [Separator::create $path.sep -orient horizontal]
            pack $sep -fill x
        }
        if { $sepopt == "both" || $sepopt == "bottom" } {
            set sep [Separator::create $botframe.sep -orient horizontal]
            pack $sep -fill x
        }
    }

    # --- status bar -------------------------------------------------------------------------
    set status   [frame $botframe.status -relief flat -borderwidth 0]
    set lab      [label $status.lab -textvariable [Widget::cget $path -textvariable]]
    set indframe [frame $status.indf -relief flat -borderwidth 0]
    $status configure -height [winfo reqheight $lab]
    place $lab      -anchor w -x 0 -rely 0.5
    place $indframe -anchor e -relx 1 -rely 0.5
    pack $status    -fill x -pady 2

    pack $botframe  -side bottom -fill x
    pack $userframe -fill both -expand yes

    set _widget($path,top)      $top
    set _widget($path,ntoolbar) 0
    set _widget($path,nindic)   0

    set menu [Widget::cget $path -menu]
    if { $menu != "" } {
        _create_menubar $path $menu
    }

    bind $path <Destroy> {MainFrame::_destroy %W}

    rename $path ::$path:cmd
    proc ::$path { cmd args } "return \[eval MainFrame::\$cmd $path \$args\]"

    return $path
}


# ------------------------------------------------------------------------------
#  Command MainFrame::configure
# ------------------------------------------------------------------------------
proc MainFrame::configure { path args } {
    variable _widget

    set res [Widget::configure $path $args]

    if { [Widget::hasChanged $path -textvariable newv] } {
        uplevel \#0 $path.botf.status.lab configure -textvariable $newv
    }
    if { [Widget::hasChanged $path -variable newv] &&
         [winfo exists $path.botf.status.progress] } {
        ProgressBar::configure $path.botf.status.progress -variable $newv
    }
    return $res
}


# ------------------------------------------------------------------------------
#  Command MainFrame::cget
# ------------------------------------------------------------------------------
proc MainFrame::cget { path option } {
    return [Widget::cget $path $option]
}


# ------------------------------------------------------------------------------
#  Command MainFrame::getframe
# ------------------------------------------------------------------------------
proc MainFrame::getframe { path } {
    return $path.user
}


# ------------------------------------------------------------------------------
#  Command MainFrame::addtoolbar
# ------------------------------------------------------------------------------
proc MainFrame::addtoolbar { path } {
    global   tcl_platform
    variable _widget

    set index     $_widget($path,ntoolbar)
    set toolframe $path.topf.f$index
    if { $tcl_platform(platform) == "unix" } {
        frame $toolframe -relief raised -borderwidth 1
    } else {
        frame $toolframe -relief flat -borderwidth 0
        set sep [Separator::create $toolframe.sep -orient horizontal]
        pack $sep -fill x
    }
    set toolbar [frame $toolframe.tb -relief flat -borderwidth 2]
    pack $toolbar -anchor w
    incr _widget($path,ntoolbar)
    grid $toolframe -column 0 -row $index -sticky ew
    return $toolbar
}


# ------------------------------------------------------------------------------
#  Command MainFrame::gettoolbar
# ------------------------------------------------------------------------------
proc MainFrame::gettoolbar { path index } {
    return $path.topf.f$index.tb
}


# ------------------------------------------------------------------------------
#  Command MainFrame::showtoolbar
# ------------------------------------------------------------------------------
proc MainFrame::showtoolbar { path index bool } {
    variable _widget

    set toolframe $path.topf.f$index
    if { [winfo exists $toolframe] } {
        if { !$bool && [llength [grid info $toolframe]] } {
            grid forget $toolframe
            $path.topf configure -height 1
        } elseif { $bool && ![llength [grid info $toolframe]] } {
            grid $toolframe -column 0 -row $index -sticky ew
        }
    }
}


# ------------------------------------------------------------------------------
#  Command MainFrame::showstatusbar
# ------------------------------------------------------------------------------
proc MainFrame::showstatusbar { path name } {
    set status $path.botf.status
    if { $name == "none" } {
        pack forget $status
    } else {
        pack $status -fill x
        switch -- $name {
            status {
                destroy $status.progress
                place $status.lab -anchor w -x 0 -rely 0.5
            }
            progression {
                set progress $status.progress
                if { ![winfo exists $progress] } {
                    ProgressBar::create $progress \
                        -width       50 \
                        -height      [expr [winfo height $status.lab]-4] \
                        -troughcolor [$status.lab cget -background] \
                        -borderwidth 2 \
                        -variable    [Widget::cget $path -variable]
                    place $progress -anchor w -x 0 -rely 0.5
                    place $status.lab -anchor w -x 54 -rely 0.5
                }
            }
        }
    }
}


# ------------------------------------------------------------------------------
#  Command MainFrame::addindicator
# ------------------------------------------------------------------------------
proc MainFrame::addindicator { path args } {
    variable _widget

    set index $_widget($path,nindic)
    set indic $path.botf.status.indf.f$index
    eval label $indic $args -relief sunken -bd 1

    pack $indic -side left -anchor w -padx 2

    incr _widget($path,nindic)

    return $indic
}


# ------------------------------------------------------------------------------
#  Command MainFrame::getindicator
# ------------------------------------------------------------------------------
proc MainFrame::getindicator { path index } {
    return $path.botf.status.indf.f$index
}


# ------------------------------------------------------------------------------
#  Command MainFrame::_destroy
# ------------------------------------------------------------------------------
proc MainFrame::_destroy { path } {
    variable _widget

    Widget::destroy $path
    $_widget($path,top) configure -menu {}
    destroy $_widget($path,top).menubar
    unset _widget($path,top)
    unset _widget($path,ntoolbar)
    unset _widget($path,nindic)
    rename $path {}
}


# ------------------------------------------------------------------------------
#  Command MainFrame::_create_menubar
# ------------------------------------------------------------------------------
proc MainFrame::_create_menubar { path descmenu } {
    variable _widget

    set top     $_widget($path,top)
    set menubar [menu $top.menubar -type menubar -tearoff 0]
    $top configure -menu $menubar

    set count 0
    foreach {name tearoff entries} $descmenu {
        set opt  [_parse_name $name]
        set menu $menubar.menu$count
        eval $menubar add cascad $opt -menu $menu
        menu $menu -tearoff $tearoff
        _create_entries $path $menu $entries
        incr count
    }
}


# ------------------------------------------------------------------------------
#  Command MainFrame::_create_entries
# ------------------------------------------------------------------------------
proc MainFrame::_create_entries { path menu entries } {
    variable _widget

    set count      [$menu cget -tearoff]
    set registered 0
    foreach entry $entries {
        set len  [llength $entry]
        set type [lindex $entry 0]
        set opt  [_parse_name [lindex $entry 1]]
        if {$type == "separator"} {
            $menu add separator
        } elseif { $type == "cascad" } {
            set submenu $menu.menu$count
            eval $menu add cascad $opt -menu $submenu
            menu $submenu -tearoff [lindex $entry 2]
            _create_entries $path $submenu [lindex $entry 3]
        } else {
            set cmd  [lindex $entry 2]
            set desc [lindex $entry 3]
            if { $cmd != "" } {
                lappend opt -command $cmd
            }
            if { $desc != "" } {
                if { !$registered } {
                    DynamicHelp::register $menu menu [Widget::cget $path -textvariable]
                    set registered 1
                }
                DynamicHelp::register $menu menuentry $count $desc
            }
            if { $type == "command" } {
                set accel [_parse_accelerator [lindex $entry 4]]
                if { [llength $accel] } {
                    lappend opt -accelerator [lindex $accel 0]
                    bind $_widget($path,top) [lindex $accel 1] "$menu invoke $count"
                }
                set useropt [lindex $entry 5]
            } else {
                set var [lindex $entry 4]
                if { $var != "" } {
                    lappend opt -variable $var
                }
                if { $type == "radiobutton" } {
                    set val [lindex $entry 5]
                    if { $val != "" } {
                        lappend opt -value $val
                    }
                    set useropt [lindex $entry 6]
                } elseif { $type == "checkbutton" } {
                    set useropt [lindex $entry 5]
                } else {
                    return -code error "invalid menu type \"$type\""
                }
            }
            eval $menu add $type $opt $useropt
        }
        incr count
    }
}


# ------------------------------------------------------------------------------
#  Command MainFrame::_parse_name
# ------------------------------------------------------------------------------
proc MainFrame::_parse_name { menuname } {
    set idx [string first "&" $menuname]
    if { $idx == -1 } {
        return [list -label $menuname]
    } else {
        set beg [string range $menuname 0 [expr $idx-1]]
        set end [string range $menuname [expr $idx+1] end]
        append beg $end
        return [list -label $beg -underline $idx]
    }
}


# ------------------------------------------------------------------------------
#  Command MainFrame::_parse_accelerator
# ------------------------------------------------------------------------------
proc MainFrame::_parse_accelerator { desc } {
    if { [llength $desc] == 2 } {
        set seq [lindex $desc 0]
        set key [lindex $desc 1]
        switch -- $seq {
            Ctrl {
                set accel "Ctrl+[string toupper $key]"
                set event "<Control-[string tolower $key]>"
            }
            Alt {
                set accel "Atl+[string toupper $key]"
                set event "<Alt-[string tolower $key]>"
            }
            CtrlAlt {
                set accel "Ctrl+Alt+[string toupper $key]"
                set event "<Control-Alt-[string tolower $key]>"
            }
            default {
                return -code error "invalid accelerator code $seq"
            }
        }
        return [list $accel $event]
    }
    return {}
}


